from __future__ import annotations

import concurrent.futures
import functools
import itertools
from pathlib import Path
from typing import Dict, Iterable, Optional, Tuple

from audiomatch import files, fingerprints
from audiomatch.constants import DEFAULT_LENGTH


def match(
    *paths: Path,
    length: int = DEFAULT_LENGTH,
    extensions: Optional[Iterable[str]] = None,
) -> Dict[Tuple[Path, Path], float]:
    """
    Finds similar audio files in paths.

    Args:
        length: specifies how many seconds of the input audio to take for analysis.
            Defaults to 120.
        extensions: Take only files with given extensions. It has no effect on paths
            that already have extension.

    Returns:
        A dictionary where key is a pair of filepaths and value is a score between them.
    """
    pairs = list(files.pair(*paths, extensions=extensions))
    filepaths = list(set(itertools.chain.from_iterable(pairs)))
    func = functools.partial(fingerprints.calc, length=length)
    with concurrent.futures.ThreadPoolExecutor() as executor:
        fps = {filepaths[i]: fp for i, fp in enumerate(executor.map(func, filepaths))}

    # Using multiprocessing.Pool.starmap method we can avoid writing wrapper to unpack
    # arguments. However, multiprocessing.Pool doesn't play nicely with coverage, and
    # require to explicitly call 'pool.join'
    with concurrent.futures.ProcessPoolExecutor() as executor:
        scores = executor.map(_compare, ((fps[a], fps[b]) for a, b in pairs))

    return dict(zip(pairs, scores))


def _compare(pair):
    """Just a wrapper for fingerprints.compare, that unpack its first argument"""
    return fingerprints.compare(*pair)



import argparse
import pathlib
import sys

from audiomatch import match, reports
from audiomatch.constants import DEFAULT_EXTENSIONS, DEFAULT_LENGTH
from audiomatch.exceptions import NotEnoughFiles


def invoke():
    parser = get_parser()
    args = parser.parse_args()

    try:
        matches = match.match(*args.path, length=args.length, extensions=args.extension)
    except NotEnoughFiles as exc:
        sys.exit(exc)

    reports.console(matches)


def get_parser():
    parser = argparse.ArgumentParser(
        prog="audiomatch",
        description="Compare audio files in a provided paths and show similar ones",
    )
    parser.add_argument(
        "path",
        type=pathlib.Path,
        nargs="+",
        help="filepath, glob-style pattern or a directory",
    )
    parser.add_argument(
        "-l",
        "--length",
        type=int,
        default=DEFAULT_LENGTH,
        metavar="SECONDS",
        help=f"""
            specifies how many seconds of the input audio to take for analysis
            (default: {DEFAULT_LENGTH})
        """,
    )
    parser.add_argument(
        "-e",
        "--extension",
        action="append",
        help=f"""
            Take only files with given extension
            (default: {', '.join(DEFAULT_EXTENSIONS)})
        """,
    )
    return parser



from __future__ import annotations

import glob
import itertools
from pathlib import Path
from typing import Iterable, Iterator, Optional, Tuple

from audiomatch.constants import DEFAULT_EXTENSIONS
from audiomatch.exceptions import NotEnoughFiles


def pair(
    *paths: Path, extensions: Optional[Iterable[str]] = None
) -> Iterator[Tuple[Path, Path]]:
    """
    Returns a cartesian product of all files found in paths.

    Args:
        *paths: a filepath, glob-style pattern or a directory.
        extensions: only take files with given extensions. It has no effect for paths
        that already have extension.

    Raises:
        ValueError: If only single file found in paths.

    Returns: A 2-length tuples where each element is a filepath.
    """
    if extensions is None:
        extensions = DEFAULT_EXTENSIONS

    files = []
    for path in paths:
        if path.is_dir():
            path = path.joinpath("*")
        files.append(
            [
                Path(pathname)
                for pathname in glob.iglob(str(path))
                if path.suffix or Path(pathname).suffix in extensions
            ]
        )

    if len(files) == 1 and len(files[0]) > 1:
        return itertools.combinations(*files, 2)
    elif len(files) > 1:
        return itertools.chain.from_iterable(
            itertools.product(*group) for group in itertools.combinations(files, 2)
        )
    else:
        raise NotEnoughFiles("Not enough input files.")



import sys


def console(matches) -> None:
    """Print similar audio to standard output"""
    similars = _join(_adjancency_list(matches))
    lines = ["\n".join(str(node) for node in similar) for similar in sorted(similars)]

    if lines:
        sys.stdout.write("These files sound similar:\n\n")
        sys.stdout.write("\n\n---\n\n".join(lines))
    else:
        sys.stdout.write("No matches found.")

    sys.stdout.write("\n\n")


def _join(graph):
    """
    Groups all adjacent nodes into a unified list.

    For example, for matches '{(a, b): 0.7, (b, c): 0.8}' result will be '[a, b, c]'
    """
    return [
        sorted(similar)
        for node, edges in graph.items()
        if (similar := _traverse(graph, node))
    ]


def _adjancency_list(matches, score=0.61):
    """Returns an adjacency list for matches"""
    graph = {}
    for pair, score in matches.items():
        if score > 0.61:
            a, b = pair
            graph.setdefault(a, []).append(b)
            graph.setdefault(b, []).append(a)
    return graph


def _traverse(graph, node, visited=[]):
    """Returns list of all direct/indirect adjacent nodes for a given node"""
    nodes = []
    if node in visited:
        return nodes
    visited.append(node)
    nodes.append(node)
    for adjacent in graph[node]:
        nodes.extend(_traverse(graph, adjacent))
    return nodes



import statistics
import subprocess

from audiomatch import popcount

CONFIDENCE_SCORE = 0.665
CORRECTION = 0.95
SCORE_MEDIAN_DELTA = 0.04


def calc(path, length=120):
    # TODO: Probably it would be better to parse json output
    fp = subprocess.run(
        ["fpcalc", "-rate", "11025", "-raw", "-length", str(length), str(path)],
        stdout=subprocess.PIPE,
    )
    if lines := fp.stdout.decode().splitlines():
        return [int(value) for value in lines[1].strip("FINGERPRINT=").split(",")]
    return []


def compare(fp1, fp2):
    # Take first 30 seconds of the the shortest fingerprint and try to find it in a
    # longer one
    if len(fp1) > len(fp2):
        return find_best_score(fp1, fp2[: seconds(30)])
    else:
        return find_best_score(fp2, fp1[: seconds(30)])


def find_best_score(fp1, fp2):
    # Fingerprints lesser than 10 seconds don't have enough data for analysis
    if len(fp1) > seconds(10) and len(fp2) > seconds(10):
        results = [correlation(_fp1, _fp2) for _fp1, _fp2 in cross(fp1, fp2)]
        score = max(results)

        # With this score we assume two fingerprints are similar
        if score >= CONFIDENCE_SCORE:
            return score

        i = results.index(score)

        # A lot of false positives happen with fingerprints less than 20 sec, so we need
        # to slightly correct the score. If it really matches another fingerprint the
        # score will be high enough to not be affected by this
        if min(len(fp1), len(fp2)) < seconds(20):
            score *= CORRECTION

        # Usually, when two fingerprints match they have a high score at a match point
        # and lesser scores before and after match point. This assumption helps us
        # avoid false positives - they tend to have rather the same score regardless
        # of how fingerprints are aligned
        offset = 5
        samples = results[i - offset : i] + results[i + 1 : i + offset + 1]
        if score - statistics.median(samples) > SCORE_MEDIAN_DELTA:
            return score

    return 0.0


def correlation(fp1, fp2):
    error = sum(popcount.popcount(x ^ y) for x, y in zip(fp1, fp2))
    return 1.0 - error / 32.0 / min(len(fp1), len(fp2))


def cross(fp1, fp2):
    length = min(len(fp1), len(fp2))
    span = min(length // 4, seconds(5))
    limit = max(len(fp1), len(fp2)) - length - span
    step = seconds(0.3)
    # No need to trim the second fingerprint, as 'zip' will trim it automatically
    for offset in range(span, 0, -step):
        yield fp2[offset:], fp1
    for offset in range(0, limit, step):
        yield fp1[offset:], fp2


def seconds(x) -> int:
    return round(x * 7)



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['audiomatch']

package_data = \
{'': ['*']}

entry_points = \
{'console_scripts': ['audiomatch = audiomatch.cli:invoke']}

setup_kwargs = {
    'name': 'audiomatch',
    'version': '0.1.5',
    'description': 'A small command-line tool to find similar audio files',
    'long_description': '==========\naudiomatch\n==========\n\n.. image:: https://github.com/unmade/audiomatch/workflows/lint%20and%20test/badge.svg?branch=master\n    :alt: Build Status\n    :target: https://github.com/unmade/audiomatch/blob/master/.github/workflows/lint-and-test.yml\n\n.. image:: https://codecov.io/gh/unmade/audiomatch/branch/master/graph/badge.svg\n    :alt: Coverage Status\n    :target: https://codecov.io/gh/unmade/audiomatch\n\n.. image:: https://img.shields.io/pypi/v/audiomatch.svg\n    :alt: PyPI Package latest release\n    :target: https://pypi.org/project/audiomatch\n\n.. image:: https://img.shields.io/badge/License-MIT-purple.svg\n    :alt: MIT License\n    :target: https://github.com/unmade/apiwrappers/blob/master/LICENSE\n\n\nA small command-line tool to find similar audio files\n\nInstallation\n============\n\nFirst, install the Chromaprint_ fingerprinting library by Lukáš Lalinský. (The library\nitself depends on an FFT library, but it\'s smart enough to use an algorithm from\nsoftware you probably already have installed; see the Chromaprint page for details)\n\nThen you can install this library:\n\n.. code-block:: bash\n\n    pip install audiomatch\n\nYou can avoid installing all this libraries on your computer and run everything in\ndocker:\n\n.. code-block:: bash\n\n    docker run --rm -v "$(pwd)":/tmp fdooch/audiomatch "/tmp/*"\n\nQuickstart\n==========\n\nSuppose, we have a directory with Nirvana songs:\n\n.. code-block:: bash\n\n    $ ls demo\n    All Apologies (In Utero).m4a           Dumb (Unplugged in NYC).m4a\n    All Apologies (Unplugged in NYC).m4a   Pennyroyal Tea (In Utero).m4a\n    Dumb (In Utero).m4a                    Pennyroyal Tea (Solo Acoustic).mp3\n    Dumb (Radio Appearance, 1991).mp3      Pennyroyal Tea (Unplugged in NYC).m4a\n\nLet\'s find out which files sound similar:\n\n.. code-block:: bash\n\n    $ audiomatch --length 300 ./demo\n    These files sound similar:\n\n    ./demo/All Apologies (In Utero).m4a\n    ./demo/All Apologies (Unplugged in NYC).m4a\n\n    ---\n\n    ./demo/Dumb (In Utero).m4a\n    ./demo/Dumb (Unplugged in NYC).m4a\n\n    ---\n\n    ./demo/Pennyroyal Tea (In Utero).m4a\n    ./demo/Pennyroyal Tea (Solo Acoustic).mp3\n    ./demo/Pennyroyal Tea (Unplugged in NYC).m4a\n\n*Note #1: input audio files should be at least 10 seconds long*\n\n*Note #2: in some rare cases false positives are possible*\n\nWhat\'s happening here is that *audiomatch* takes all audio files from the directory and\ncompares them with each other.\n\nYou can also compare file with another file, file and directory, or directory to\ndirectory. If you need to, you can provide glob-style patterns, but don\'t forget to\nquote it, because otherwise shell expanded it for you. For example, let\'s compare all\n``.mp3`` files with ``.m4a`` files:\n\n.. code-block:: bash\n\n    $ audiomatch  "./demo/*.mp3" "./demo/*.m4a"\n    These files sound similar:\n\n    ../demo/Pennyroyal Tea (Solo Acoustic).mp3\n    ../demo/Pennyroyal Tea (Unplugged in NYC).m4a\n\nThis time, *audiomatch* took all files with ``.mp3`` extension and compare them with\nall files with ``.m4a`` extension.\n\nNote, how there is no In Utero version in the output. The reason it is present in the\nprevious output, because it actually similar with Unplugged version and then transitive\nlaw applies: if ``a = b`` and ``b = c``, then ``a = c``.\n\n--length\n--------\n\nThe ``--length`` specifies how many seconds to take for analysis from the song. Default\nvalue is 120 and it is good enough to find exactly the same song, but maybe in different\nquality. However, for a more complicated cases like same song played in different tempo\nthe more input we have the more accurate results are.\n\n--extension\n-----------\n\nBy default, ``audiomatch`` looks for files with ``.m4a``, ``mp3``, ``.caf`` extensions.\nIn theory, audio formats supported by ffmpeg_ also supported by *audiomatch*. You can\ntell to *audiomatch* to look for a specific format by using ``--extension`` flag:\n\n.. code-block:: bash\n\n    $ audiomatch -e .ogg -e .wav ./demo\n    Not enough input files.\n\nIndeed, we tried to compare files with ``.ogg`` and ``.wav`` extension, but there are\nno such files in the demo directory.\n\nMotivation\n==========\n\nI play guitar and do recordings from time to time mainly with Voice Memos on iPhone.\nOver the years, I have hundreds of recordings like that and I though it would be cool\nto find all the similar ones and see how I progress over the years.\n\nThat\'s why I wrote this library.\n\nReferences\n==========\n\n- Chromaprint_ and pyacoustid_ libraries\n- `Example: How to compare fingerprints`_\n- `Example: How to compare shifted fingerprints`_ (note: the code is a little bit weird)\n- `Explanation: How to compare fingerprints`_\n- `Popcount in Python with benchmarks`_\n\n.. _Chromaprint: https://github.com/acoustid/chromaprint\n.. _`Example: How to compare fingerprints`: https://gist.github.com/lalinsky/1132166\n.. _`Example: How to compare shifted fingerprints`: https://medium.com/@shivama205/audio-signals-comparison-23e431ed2207\n.. _`Explanation: How to compare fingerprints`: https://groups.google.com/forum/#!msg/acoustid/Uq_ASjaq3bw/kLreyQgxKmgJ\n.. _ffmpeg: http://ffmpeg.org\n.. _`Popcount in Python with benchmarks`: http://www.valuedlessons.com/2009/01/popcount-in-python-with-benchmarks.html\n.. _`pyacoustid`: https://github.com/beetbox/pyacoustid\n',
    'author': 'Aleksei Maslakov',
    'author_email': 'lesha.maslakov@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



"""
Source:
    http://www.valuedlessons.com/2009/01/popcount-in-python-with-benchmarks.html
"""


def popcount_table16():
    table = [0] * 2 ** 16
    for i in range(len(table)):
        table[i] = (i & 1) + table[i >> 1]
    return table


POPCOUNT_TABLE16 = popcount_table16()


def popcount(x):
    return POPCOUNT_TABLE16[x & 0xFFFF] + POPCOUNT_TABLE16[(x >> 16) & 0xFFFF]
